/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.data.fields;

import amberstone.data.*;
import amberstone.lang.*;

public class Bytes extends Field implements Key
{
	private int maxLength;
	private int strength;
	private int length;
	private int hashCode;
	private ByteArray writeBytes = new ByteArray();
	private ByteArray readBytes = new ByteArray();
	private BytesFunction loadFunc;
	private byte[] fill = new byte[16];
	private boolean positioned;
	private boolean saveHash;

	public Bytes(int type,
				 String name,
				 int index,
				 int maxLength,
				 int strength,
				 boolean saveHash,
				 BytesFunction loadFunc)
	{
		super(type, name, index);
		this.saveHash = saveHash;
		this.maxLength = maxLength;
		this.strength = strength;
		this.loadFunc = loadFunc;
	}

	public void clear()
	{
	}

	public boolean equalsKey(Key k)
	{
		if(this.hashCode == k.hashCode())
		{
			ByteArray compare = k.byteValue();
			int compareLength = compare.length;
			if(readBytes.length == compareLength)
			{
				byte[] compBuf = compare.getBuffer();
				int j = compare.offset;
				int max = j+compareLength;

				byte[] buf = readBytes.getBuffer();
				int i = readBytes.offset;
				int s = 5;

				while( j < max && buf[i] == compBuf[j])
				{
					i+=s;
					j+=s;
				}

				if(j >= max)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}

	public void setLoadFunction(BytesFunction loadFunc)
	{
		this.loadFunc = loadFunc;
	}

	public Field cloneField()
	{
		return new Bytes(this.type, this.name, this.columnIndex, this.maxLength, this.strength, this.saveHash, this.loadFunc);
	}

	public boolean copyKey(Key key)
	{
		return false;
	}

	public Key getKey()
	{
		return this;
	}

	public int getType()
	{
		return this.type;
	}

	public int intValue()
	{
		return 0;
	}

	public long longValue()
	{
		return 0L;
	}

	public Key cloneKey()
	{
		BytesKey klone = new BytesKey(maxLength);
		klone.setType(this.type);
		klone.setStrength(strength);
		klone.copyKey(this);
		return klone;
	}

	public boolean isNull()
	{
		return this.readBytes.length == 0;
	}

	public ByteArray byteValue()
	{
		return this.readBytes;
	}

	public ByteArray getBytes(Function caller)
	{
		return this.readBytes;
	}

	public int getTime(Function caller) throws FunctionException
	{
		throw new FunctionException("Time not supported for bytes type.", caller);
	}

	public int getInt(Function caller) throws FunctionException
	{
		return this.readBytes.getLength();
	}

	public long getLong(Function caller)
	{
		return this.readBytes.getLength();
	}

	public int maxLength()
	{
		return 1+this.maxLength;
	}

	public int minLength()
	{
		return 1;
	}

	public int prepareWrite() throws FieldException
	{
		try
		{
			ByteArray rawBytes = this.loadFunc.getBytes();
			writeBytes.set(rawBytes.getBuffer(), rawBytes.getOffset(), rawBytes.getLength());
		}
		catch(Exception e)
		{
			throw new FieldException(this.name, e);
		}

		int fl = saveHash ? 4 : 0;

		if(writeBytes.length > 127)
		{
			fl += 2+writeBytes.length;
		}
		else
		{
			fl += 1+writeBytes.length;
		}

		if(buffer.getPointer()+fl <  buffer.getBuffer().length)
		{
			return fl;
		}
		else
		{
			return -1;
		}
	}

	public void init(int recordsPerBlock)
	{
		byte[] buf = new byte[(maxLength()*recordsPerBlock)];
		buffer.set(buf, 0, buf.length);
	}

	public int writeField()
	{
		if(writeBytes.length > maxLength)
		{
			writeBytes.length = maxLength;
		}

		buffer.writeBytes(writeBytes, this.maxLength);
		buffer.writeInt1(0);
		return writeBytes.length+1;
	}

	public void readField()
	{
		int h = 0;
		int l = 0;

		if(buffer.length == 0)
		{
			readBytes.set(buffer.getBuffer(), 0, 0);
			readBytes.setHash(0);
			return;
		}

		int pointer = buffer.getPointer();
		byte[] raw = buffer.getBuffer();

		while(true)
		{
			int b = raw[pointer++];
			if(b == 0)
			{
				this.readBytes.set(buffer.getBuffer(), buffer.getPointer(), l);
				this.hashCode = h & 0x7FFFFFFF;
				this.readBytes.setHash(this.hashCode);
				buffer.seek(pointer);
				break;
			}
			else
			{
				++l;
				h = (31*h) + b;
			}
		}
	}

	public int hashCode()
	{
		return this.hashCode;
	}

	public String signiture()
	{
		if(saveHash)
		{
			return this.name+".hash";
		}
		else
		{
			return this.name;
		}
	}

	public String toString()
	{
		return readBytes.toString();
	}
}