/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.lang;

import java.util.List;
import java.util.ArrayList;
import amberstone.system.Logger;
import amberstone.system.Config;

public abstract class Function
{
	private Function caller;
	private String callName;
	private int tokenIndex;
	private List<String> tokenList;
	protected Function[] funcs = new Function[0];
	protected boolean debug;

	public void setCallName(String callName)
	{
		this.callName = callName;
	}

	public void setDebug(boolean debug)
	{
		this.debug = debug;
	}

	public String getCallName()
	{
		return this.callName;
	}

	public List<String> getTokenList()
	{
		return this.tokenList;
	}

	public void setCaller(Function caller)
	{
		this.caller = caller;
	}

	public Function getCaller()
	{
		return this.caller;
	}

	public void setTokenIndex(int tokenIndex)
	{
		this.tokenIndex = tokenIndex;
	}

	public int getTokenIndex()
	{
		return this.tokenIndex;
	}

	public void _init(Config config, Logger logger) throws FunctionException
	{

		for(int i=0; i<funcs.length; i++)
		{
			funcs[i]._init(config, logger);
		}

		init(config, logger);

	}

	public void init(Config config, Logger logger) throws FunctionException
	{

	}

	public void _destroy() throws FunctionException
	{
		for(int i=0; i<funcs.length; i++)
		{
			funcs[i]._destroy();
		}

		destroy();
	}

	public void destroy() throws FunctionException
	{

	}

	public String getCallTrace()
	{
		List<String> calls = new ArrayList();
		calls.add(callName);
		Function f = caller;

		while(f != null)
		{
			calls.add(f.getCallName());
			f = f.getCaller();
		}

		StringBuilder callBuf = new StringBuilder();

		int comma = 0;
		for(int i=calls.size()-1; i>=0; i--)
		{
			if(comma>0)
			{
				callBuf.append("->");
			}
			comma++;
			callBuf.append(calls.get(i)+"()");
		}

		return callBuf.toString();
	}

	public void link(Object[] params) throws CompilationException
	{

		this.funcs = new Function[params.length];

		for(int i=0; i<params.length; i++)
		{
			this.funcs[i] = (Function)params[i];
			this.funcs[i].setCaller(this);
		}

		FunctionType[][] allowGroup = strict();

		boolean lengthGood = false;
		boolean allowed = false;

		if(allowGroup != null)
		{
			if(allowGroup.length == 0 && funcs.length == 0)
			{
				lengthGood = true;
				allowed = true;
			}

			for(int i=0; i<allowGroup.length; i++)
			{
				FunctionType[] allow = allowGroup[i];
				if(funcs.length == allow.length)
				{
					lengthGood = true;
				}
			}

			if(!lengthGood)
			{
				throw new CompilationException(tokenIndex, new UsageException("Invalid parameter count:"+funcs.length, this));
			}

			for(int d=0; d < allowGroup.length; d++)
			{
				FunctionType[] allow = allowGroup[d];
				if(functionsMatch(funcs, allow))
				{
					allowed = true;
					break;
				}
			}

			if(!allowed)
			{
				throw new CompilationException(tokenIndex+2, new UsageException("Invalid parameters for function: "+getCallName(), this));
			}
		}

		params(funcs);
	}

	private boolean functionsMatch(Function[] funcs, FunctionType[] allow)
	{
		if(funcs.length != allow.length)
		{
			return false;
		}

		for(int i=0; i<funcs.length; i++)
		{
			Class funcClass = funcs[i].getClass();
			String allowedFunc = allow[i].getClassName();
			if(!functionMatch(funcClass, allowedFunc))
			{
				return false;
			}
		}

		return true;
	}

	private boolean functionMatch(Class funcClass, String allowedFunc)
	{
		//System.out.println("Looking for:"+ allowedFunc);

		Class c = funcClass;
		List<String> compareTo = new ArrayList();
		while(c != null)
		{
			String name = c.getName();
			compareTo.add(name);
			Class[] interfaces = c.getInterfaces();
			for(int n=0; n < interfaces.length; n++)
			{
				String iname = interfaces[n].getName();
				compareTo.add(iname);
				Class[] si = interfaces[n].getInterfaces();
				for(int i=0; i<si.length; i++)
				{
					compareTo.add(si[i].getName());
				}
			}

			c = c.getSuperclass();
		}

		for(int n=0; n < compareTo.size(); n++)
		{
			String iname = compareTo.get(n);
			//System.out.println("Interface:"+iname);
			if(iname.equals(allowedFunc))
			{
				return true;
			}
		}

		return false;
	}

	public String getReturnType(String iname)
	{
		for(FunctionType ft : FunctionType.values())
		{
			if(ft.getClassName().equals(iname))
			{
				return ft.toString();
			}
		}

		return null;
	}

	public abstract FunctionType[][] strict();


	public void params(Function[] params) throws ParameterException
	{

	}
}