/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.load;

import java.io.File;
import java.io.FileFilter;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

import amberstone.lang.*;
import amberstone.system.*;
import amberstone.read.DirDesc;
import amberstone.dictionary.Dictionary;
import amberstone.dictionary.DiskDictionary;
import amberstone.file.FileHandler;
import amberstone.file.FileStat;
import amberstone.file.FilePath;
import amberstone.data.DataRecord;
import amberstone.data.Field;

import java.util.Arrays;

public class Loader
{
	private Config config;
	private IndexConfig indexConfig;
	private String indexDir;
	private String loadQueue;
	private FileHandler fileHandler = new FileHandler();

	private byte delimiter;
	private DelimitedRecord delimitedRecord;
	private int maxFileSize;
	private Logger logger;
	private int logLevel;
	private int exceptions;
	private	int totalLoaded = 0;

	private int maxExceptions;
	public static volatile boolean go = true;
	public static volatile boolean running;

	public Loader(Config config, boolean archive) throws Exception
	{
		logLevel = config.getLogLevel();
		int logMode = config.getLogMode();
		String logsHome = config.getLogsHome();
		logger = new Logger(logsHome, "loader", logMode);
		logger.timestamp();
		logger.log("Loader startup");

		this.config = config;

		try
		{
			this.delimiter = (byte)config.getProperty(Config.LOAD_DELIM).charAt(1);
			this.maxFileSize = Integer.parseInt(config.getProperty(Config.INDEX_BLOCK_SIZE));
			this.maxExceptions = Integer.parseInt(config.getProperty(Config.LOAD_EXCEPTIONS_MAX));

			indexConfig = config.getIndexConfig();
			this.delimitedRecord =     indexConfig.getDelimitedRecord();
			this.loadQueue = indexConfig.getLoadDir();
			this.indexDir = indexConfig.getIndexDir() ;

			logger.ln();
			logger.log("Loader settings");
			logger.log("-"+Config.LOAD_DELIM+": "+config.getProperty(Config.LOAD_DELIM));
			logger.log("-"+Config.INDEX_BLOCK_SIZE+": "+this.maxFileSize);
			logger.log("-Load Dir: "+ this.loadQueue);
			logger.log("-Index Dir: "+ this.indexDir);
		}
		catch(Exception e)
		{
			logger.log("The following error occured while initializing the loader:");
			logger.log(e);
			throw e;
		}
		finally
		{

		}
	}

	private String pad(int i)
	{
		String s = Integer.toString(i);
		if(s.length() == 1)
		{
			return 0+s;
		}

		return s;
	}

	public void load() throws Exception
	{
		DataRecordWriter dataRecordWriter = null;
		logger.ln();
		logger.timestamp();
		logger.log("Starting load");
		logger.ln();

		DataRecord dataRecord = indexConfig.getDataRecord();

		try
		{

			dataRecord.initWrite(maxFileSize);

			dataRecordWriter = new DataRecordWriter(maxFileSize, dataRecord.maxLength(), this.indexDir);

			File rootDir = new File(this.loadQueue);
			FileStat[] topLevelFiles = fileHandler.getFiles(rootDir, null, false); // get the top level files but don't recurse

			if(topLevelFiles.length > 0)
			{
				sweep(topLevelFiles, this.indexDir);

				writeFiles(topLevelFiles,
						   dataRecord,
						   dataRecordWriter); // Write the top level files

			}

			//Load directories
			File[] years = fileHandler.getDirs(rootDir);
			for(int y=0; y<years.length; y++)
			{
				File[] months = fileHandler.getDirs(years[y]);
				for(int m=0; m<months.length; m++)
				{
					File[] days = fileHandler.getDirs(months[m]);
					for(int d=0; d<days.length; d++)
					{
						if(go)
						{
							FileStat[] lfiles = fileHandler.getFiles(days[d], null, false);
							sweep(lfiles, this.indexDir);
							writeFiles(lfiles, dataRecord, dataRecordWriter);
						}
					}
				}
			}

			logger.ln();
			logger.log("Total records loaded: "+totalLoaded);
			logger.log("Total errors: "+exceptions);
			logger.ln();
			logger.timestamp();
			logger.log("Finished load");

			System.out.println("");
			System.out.println("Total records loaded: "+totalLoaded);
			System.out.println("Total errors: "+exceptions);
		}
		catch(LoadException se)
		{
			se.log(logLevel, logger);
			logger.ln();
			logger.log("Load aborted due to error(s)");
			throw se;
		}
		catch(Exception e)
		{
			logger.log("Exception during load.");
			logger.log("Load aborted due to error(s)");
			logger.log(e);
			throw e;
		}
		finally
		{
			try
			{
				dataRecordWriter.close(dataRecord);
			}
			catch(Exception e1)
			{

			}

			logger.close();
		}
	}

	private void sweep(FileStat[] files, String dataWarehouse) throws Exception
	{
		Map fullFiles = new HashMap();
		for(int i=0; i<files.length; i++)
		{
			FileStat f = files[i];
			int yyyy = (f.name / 10000);
			int mm = (int)((f.name % (yyyy*10000)) / 100);
			int dd = f.name%((yyyy*10000)+(mm*100));

			int partNumber = f.partNumber;
			File d = new File(dataWarehouse+File.separatorChar+yyyy+File.separatorChar+pad(mm)+File.separatorChar+pad(dd));
			if(!d.exists())
			{
				continue;
			}

			File[] blocks = d.listFiles();
			if(blocks.length > 0)
			{
				f.warehouseNumber = blocks.length+1;
			}
		}

		logger.ln();
	}

	class NameFilter implements FileFilter
	{
		private int name;

		public NameFilter(int name)
		{
			this.name = name;
		}

		public boolean accept(File file)
		{
			String[] parts = file.getName().split("\\.");
			int n = Integer.parseInt(parts[0]);
			if(name == n)
			{
				return true;
			}
			else
			{
				return false;
			}
		}

	}

	public void writeFiles(FileStat[] files,
						   DataRecord dataRecord,
						   DataRecordWriter writer)
		throws Exception
	{

		DelimitedRecordReader inFile = new DelimitedRecordReader(this.delimiter, this.delimitedRecord);
		BooleanFunction filter = config.getIndexConfig().getLoadFilter();
		Meter meter = config.getIndexConfig().getLoadMeter();
		int valveValue = 0;
		long begin = 0L;
		long end = 0L;
		running = true;


		FILE_LOOP:
		for(int f=0; (f<files.length && go); f++)
		{
			FileStat fileStat = files[f];
			logger.log("Loading file: "+fileStat.fullName);
			logger.log("File size: "+fileStat.fileLength);
			int fileExceptions = 0;

			int recordNum = 0;
			int loaded = 0;
			try
			{
				begin = System.nanoTime();
				inFile.open(fileStat.file);
				int YYYY = (int)fileStat.name/10000;
				int MM   = (int)((fileStat.name % (YYYY*10000)) / 100);
				int DD   = fileStat.name % ((YYYY*10000)+(MM*100));
				//System.out.println("Loading file: "+fileStat.fullName);
				while(inFile.readRecord() && go)
				{
					LineNumber.LINE_NUMBER = ++recordNum;
					if(filter.getBoolean())
					{
						try
						{
							meter.update();
							writer.writeData(YYYY, MM, DD, fileStat.warehouseNumber, dataRecord);
							++loaded;
						}
						catch(Exception e)
						{
							++exceptions;
							++fileExceptions;
							if(exceptions < maxExceptions)
							{
								System.out.println("Error in file:"+fileStat.fullName);
								System.out.println("Record number:"+recordNum);
								LoadException le = new LoadException(recordNum, fileStat.fullName, e);
								le.log(logLevel, logger);
							}
							else
							{
								throw e;
							}
						}
					}
				}

				totalLoaded+=loaded;
				end = System.nanoTime();

				logger.log("File loaded in: "+Double.toString((double)(end-begin)/(double)1000000));
				logger.log("Records loaded: "+loaded);
				logger.log("Errors: "+fileExceptions);
				logger.ln();
			}
			catch(Exception e)
			{
				running = false;
				throw new LoadException(recordNum, fileStat.fullName, e);
			}
			finally
			{
				inFile.close();
			}
		}

		running = false;
	}
}