/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/


package amberstone.summary;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.List;
import java.util.ArrayList;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;

import amberstone.read.*;
import amberstone.system.Logger;
import amberstone.file.FileHandler;
import amberstone.file.FileStat;
import amberstone.data.*;

import java.util.Comparator;
import java.util.Arrays;

public class ReadAheadThread
	extends Thread
{
	private String rootDir;
	private BlockingQueue<DataRecord> readyFiles;
	private BlockingQueue<DataRecord> finishedFiles;
	private int readAhead = 2;
	private int logLevel;
	private int logMode;
	private String logsHome;
	private int startYYYYMMDD;
	private int endYYYYMMDD;
	private int startBlock;

	public ReadAheadThread(String rootDir,
						   int startYYYYMMDD,
						   int endYYYYMMDD,
						   int startBlock,
						   int blockSize,
					       DataRecord dataRecord,
					       List<Field> activeFields)
	{
		this.rootDir = rootDir;
		this.startYYYYMMDD = startYYYYMMDD;
		this.endYYYYMMDD   = endYYYYMMDD;
		this.startBlock = startBlock;
		readyFiles = new ArrayBlockingQueue(readAhead+1);
		finishedFiles = new ArrayBlockingQueue(readAhead+1);

		for(int i=0; i<readAhead; i++)
		{
			DataRecord clone = dataRecord.cloneRecord();
			List<Field> activeClones = new ArrayList();
			Field[] allClones = clone.getFields();

			for(int c=0; c<allClones.length; c++)
			{
				if(activeFields.contains(allClones[c]))
				{
					activeClones.add(allClones[c]);
				}
			}

			clone.initRead(blockSize);
			clone.setActiveFields((Field[])activeClones.toArray(new Field[activeClones.size()]));
			finishedFiles.add(clone);
		}
	}

	public void setLogInfo(String logsHome, int logMode, int logLevel)
	{
		this.logsHome = logsHome;
		this.logMode = logMode;
		this.logLevel = logLevel;
	}

	public DataRecord getReadyRecord()
		throws Exception
	{
		DataRecord r = readyFiles.take();

		if(r.getReadBytes() != null)
		{
			return r;
		}
		else
		{
			return null;
		}
	}

	public void putFinishedRecord(DataRecord r)
	{
		finishedFiles.add(r);
	}

	public void run()
	{
		Logger logger = null;

		try
		{
			long begin = 0l;
			long end = 0l;

			logger = new Logger(logsHome, "read_ahead", logMode);
			logger.log("Read ahead thread starting");
			File root = new File(this.rootDir);
			File[] years = root.listFiles();
			Comparator comp = new DirAsc();
			Arrays.sort(years, comp);

			for(int y=0; y<years.length; y++)
			{
				int year = Integer.parseInt(years[y].getName());
				File[] months = years[y].listFiles();
				Arrays.sort(months, comp);
				for(int m=0; m<months.length; m++)
				{
					int month = Integer.parseInt(months[m].getName());
					File[] days = months[m].listFiles();
					Arrays.sort(days, comp);

					for(int d=0; d<days.length; d++)
					{
						int day = Integer.parseInt(days[d].getName());
						int YYYYMMDD = (year*10000)+(month*100)+day;
						if(YYYYMMDD >= startYYYYMMDD && YYYYMMDD <= endYYYYMMDD)
						{
							File[] blocks = days[d].listFiles();
							Arrays.sort(blocks, comp);
							int s = 0;
							if(YYYYMMDD == startYYYYMMDD)
							{
								if(startBlock > 0)
								{
									s = startBlock;
								}
								else if(startBlock < 0)
								{
									s = blocks.length+startBlock;
								}
							}

							for(int i=s; i<blocks.length; i++)
							{
								DataRecord readyRecord = finishedFiles.take();
								if(readyRecord.getReadBytes() == null)
								{
									readyFiles.add(readyRecord);
									return;
								}

								File block = blocks[i];
								logger.ln();
								logger.log("Reading block:"+block.getPath());

								begin = System.nanoTime();

								readyRecord.setDir(block);
								readyRecord.readBlock();
								readyFiles.add(readyRecord);
								end = System.nanoTime();
								logger.log("Time: "+Double.toString((double)(end-begin)/(double)1000000));
							}
						}
					}
				}
			}

			readyFiles.add(new DataRecord());
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			logger.close();
		}
	}
}