/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.summary;

import amberstone.system.*;
import amberstone.transaction.TransactionTrackerTemplate;
import amberstone.lang.*;
import amberstone.data.DataRecord;
import amberstone.data.ByteArray;
import amberstone.data.Field;
import amberstone.file.FileStat;

import java.io.File;
import java.util.List;
import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ArrayBlockingQueue;

public class SummaryThread extends Thread
{
	private Config config;
	DataRecord record;
	private FileTemplate[] fileTemplates;
	private FileNode[] fileNodes;
	private BlockingQueue<DataRecord> readyFile;
	private BlockingQueue<DataRecord> finishedFile;
	private int threadNumber;
	private Meter meter;
	private BooleanFunction threadFilter;
	private LongFunction grain;
	private	long currentGrain = 0L;

	public SummaryThread(Config config, DataRecord record, FileTemplate[] fileTemplates)
	{

		this.fileTemplates = fileTemplates;
		this.record = record;
		this.config = config;
		this.readyFile = new ArrayBlockingQueue(1);
		this.finishedFile = new ArrayBlockingQueue(1);
	}

	public void setGrain(LongFunction grain)
	{
		this.grain = grain;
	}

	public void setThreadNumber(int threadNumber)
	{
		this.threadNumber = threadNumber;
	}

	public FileTemplate[] getFileTemplates()
	{
		return this.fileTemplates;
	}

	public void setMeter(Meter meter)
	{
		this.meter = meter;
	}

	public void setFilter(BooleanFunction threadFilter)
	{
		this.threadFilter = threadFilter;
	}

	public void putDataRecord(DataRecord newRecord)
		throws Exception
	{
		this.readyFile.put(newRecord);
	}

	public DataRecord getDataRecord()
		throws Exception
	{
		return this.finishedFile.take();
	}

	public void run()
	{
		ByteArray bytes = new ByteArray();
		int recordLength = 0;
		int fileLength = 0;
		FileStat currentFile = null;
		int currentDay = 0;
		Logger logger = null;

		long begin = 0l;
		long end;
		int logLevel = 0;
		try
		{
			logLevel            = Integer.parseInt(config.getProperty(Config.LOG_LEVEL));
		    int logMode         = Integer.parseInt(config.getProperty(Config.LOG_MODE));
			String logsHome     = config.getLogsHome();

			logger = new Logger(logsHome, "aggregator_thread."+threadNumber, logMode);
			logger.timestamp();
			logger.log("Thread number: "+ threadNumber + "  startup");


			Field[] activeFields = record.getActiveFields();
			logger.ln();
			logger.log("Active fields");
			for(int i=0; i<activeFields.length; i++)
			{
				logger.log("-Field: "+activeFields[i].getName());
			}

			int loop = fileTemplates.length;

			while(true)
			{
				DataRecord newRecord = readyFile.take();

				if(newRecord.getReadBytes() == null)
				{
					break;
				}

				record.getData(newRecord);

				logger.ln();
				logger.log("Aggregating block: "+newRecord.getDir().getPath());

				begin = System.nanoTime();

				while(record.readRecord())
				{
					long testGrain = grain.getLong();
					if(currentGrain != testGrain)
					{
						if(fileNodes != null)
						{
							for(int i=0; i<fileTemplates.length; i++)
							{
								fileNodes[i].finish(currentGrain);
								WriterThread.put(fileNodes[i]);
								fileNodes[i] = fileTemplates[i].getFileNode();
							}
						}
						else
						{
							fileNodes = new FileNode[fileTemplates.length];
							for(int i=0; i<fileTemplates.length; i++)
							{
								fileNodes[i] = fileTemplates[i].getFileNode();
							}
						}

						currentGrain = testGrain;
					}

					for(int i=0; i<loop; i++)
					{
						fileNodes[i].update();
					}
				}

				finishedFile.put(newRecord);
				end = System.nanoTime();
				logger.log("Time: "+Double.toString((double)(end-begin)/(double)1000000));
			}
		}
		catch(SystemException se)
		{
			se.log(logLevel, logger);
			try
			{
				DataRecord dr = new DataRecord();
				dr.e = se;
				finishedFile.put(dr);
			}
			catch(Exception e1)
			{
				e1.printStackTrace();
			}
		}
		catch(Exception e)
		{
			logger.log(e);
			try
			{
				DataRecord dr = new DataRecord();
				dr.e = e;
				finishedFile.put(dr);
			}
			catch(Exception e1)
			{
				e1.printStackTrace();
			}
		}
		finally
		{
			logger.ln();
			logger.timestamp();
			logger.log("Thread number: "+ threadNumber + "  shutdown");
			logger.close();
		}
	}

	public void finish() throws FunctionException
	{
		for(int i=0; i<fileNodes.length; i++)
		{
			fileNodes[i].finish(currentGrain);
			WriterThread.put(fileNodes[i]);
		}
	}
}