/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.system;

import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Properties;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.math.BigDecimal;

import java.util.jar.*;
import java.util.zip.*;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

import amberstone.data.*;
import amberstone.data.fields.*;
import amberstone.lang.*;
import amberstone.summary.*;
import amberstone.util.KeyValuePair;
import amberstone.file.FilePath;
import amberstone.load.*;
import amberstone.dictionary.*;
import amberstone.report.*;
import amberstone.read.*;
import amberstone.time.*;
import amberstone.resources.*;
import amberstone.transaction.*;

public class Config
	extends DefaultHandler implements ErrorHandler
{
	public static String AMBERSTONE_HOME;

	public static final String DICTIONARIES_COMMIT_EVERY     = "dictionaries-commit-every";
	public static final String DICTIONARIES_TRANSACTIONAL    = "dictionaries-transactional";
	public static final String LOG_LEVEL                     = "log-level";
	public static final String LOG_MODE                      = "log-mode";
	public static final String LOAD_DELIM                    = "load-delim";
	public static final String LOAD_EXCEPTIONS_MAX           = "max-load-exceptions";
	public static final String LOAD_READ_BUFFER_SIZE         = "load-read-buffer-size";
	public static final String MAX_LOAD_FIELDS               = "max-load-fields";
	public static final String INDEX_BLOCK_SIZE              = "index-block-size";
	public static final String MAX_EXPORT_FILE_SIZE          = "export-file-size";


	private amberstone.lang.Compiler compiler = new amberstone.lang.Compiler();

	private DelimitedRecord delimitedRecord;
	private DataRecord dataRecord;
	private List fields = new ArrayList();
	private Properties props = new Properties() ;
	private Properties lang  = new Properties();
	private Properties macros = new Properties();

	private Map<String, Integer> types = new HashMap();

	private int logLevel;
	private String logsHome;
	private int logMode;

	private String pluginsHome;
	private boolean showLocation = true;

	private Map<String, Resource> resources = new HashMap();

	private SummaryThread currentSummaryThread;
	private List<SummaryThread> summaryThreads = new ArrayList();

	private List<PropTemplate> tprops = new ArrayList();

	private FileTemplate currentFileTemplate;
	private int fieldIndex = -1;
	private int reportIndex = 0;

	private Meter loadMeter;
	private Meter readMeter;

	private Stack stack = new Stack();
	private String reportName;
	private Map<String, List> reports = new HashMap();
	private Map<String, List> alerts = new HashMap();

	private Command command;

	private Map<String, String> summaries = new HashMap();

	private TreeSet activeDelimitedFields = new TreeSet();
	private TreeSet activeFields;

	private Locator locator;

	private boolean summarizeInc;


	private boolean files;
	private int threadNum;

	private boolean transactions;

	private amberstone.read.Selector select;

	private Map<String, Dictionary> dictionaries = new HashMap();
	private String dictionaryDir;

	private String mode;
	private int commitEvery;

	private BooleanFunction loadFilter;
	private BooleanFunction threadFilter;

	private TransactionTrackerTemplate currentTransaction;

	private AmberstoneConfig amberstoneConfig;
	private IndexConfig indexConfig;
	private SummarizersConfig summarizersConfig;

	public SummarizersConfig getSummarizersConfig()
	{
		return this.summarizersConfig;
	}

	public void setDocumentLocator(Locator locator)
	{
		this.locator = locator;
	}

	public int getLogLevel()
	{
		return this.logLevel;
	}

	public String getLogsHome()
	{
		return this.logsHome;
	}

	public int getLogMode()
	{
		return this.logMode;
	}

	public SummaryThread[] getSummaryThreads()
	{
		return (SummaryThread[])this.summaryThreads.toArray(new SummaryThread[this.summaryThreads.size()]);
	}

	public BooleanFunction getLoadFilter()
	{
		return this.loadFilter;
	}

	public Meter getLoadMeter()
	{
		return this.loadMeter;
	}

	public DelimitedRecord getDelimitedRecord()
	{
		return this.delimitedRecord;
	}

	public DataRecord getDataRecord()
	{
		return this.dataRecord;
	}

	public IndexConfig getIndexConfig()
	{
		return this.indexConfig;
	}

	public Map<String, Dictionary> getDictionaries()
	{
		return indexConfig.getDictionaries();
	}

	public Map<String, Resource> getResources()
	{
		return amberstoneConfig.getResources();
	}

	public String getProperty(String key) throws IOException
	{
		return this.props.getProperty(key);
	}

	public Config()
		throws Exception
	{

		AMBERSTONE_HOME = System.getenv("AMBERSTONE_HOME");

		File confDir = new File(AMBERSTONE_HOME+File.separatorChar+"conf");
		File sysini = new File(confDir, "system.ini");

		loadProps(sysini, this.props);
		verifySystem();

		File langini = new File(confDir, "language.ini");
		loadProps(langini, this.lang);

		compiler.setLang(this.lang);
		compiler.setMacros(this.macros);

		this.logsHome       = AMBERSTONE_HOME+File.separatorChar+"logs";
		this.logLevel       = Integer.parseInt(this.props.getProperty(LOG_LEVEL));
		this.logMode        = Integer.parseInt(this.props.getProperty(LOG_MODE));

		if(this.logLevel < 1 || this.logLevel > 3)
		{
			System.out.println("-Log level must either be 1, 2 or 3");
			System.exit(1);
		}

		if(this.logMode < 1 || this.logMode > 3)
		{
			System.out.println("-Log mode must either be 1, 2 or 3");
			System.exit(1);
		}
	}

	private void loadProps(File f, Properties p) throws Exception
	{
		BufferedReader in = null;
		try
		{
			in = new BufferedReader(new FileReader(f));
			for(String line = in.readLine(); line != null; line = in.readLine())
			{
				line = line.trim();
				if(line.indexOf(":") > -1 && line.indexOf("#") != 0)
				{
					int delimIndex = line.indexOf(":");
					String name  = line.substring(0, delimIndex).trim();
					String value = line.substring(delimIndex+1).trim();
					if(p.containsKey(name))
					{
						throw new Exception("Duplicate property:"+name);
					}

					p.put(name, value);
				}
			}
		}
		finally
		{
			in.close();
		}
	}

	private boolean validWorkingDir(String d)
	{
		File dir = new File(d);
		File[] files = dir.listFiles();
		boolean warehouse = false;
		boolean conf = false;
		boolean logs = false;
		boolean lib  = false;

		for(int i=0; i<files.length; i++)
		{
			String name = files[i].getName();
			if(name.equals("conf"))
			{
				conf = true;
			}
			else if(name.equals("lib"))
			{
				lib = true;
			}
			else if(name.equals("warehouse"))
			{
				warehouse = true;
			}
			else if(name.equals("logs"))
			{
				logs = true;
			}
		}

		if(warehouse && conf &&  logs && lib)
		{
			return true;
		}
		else
		{
			return false;

		}
	}

	public void parse(Command command) throws Exception
	{
		Runtime runtime = Runtime.getRuntime();

		if(command.needsResources())
		{
			runtime.addShutdownHook(new CloseResources(this));
		}

		this.amberstoneConfig = new AmberstoneConfig(command, props, lang);
		if(command instanceof LoadCommand || command instanceof ReadCommand || command instanceof SummarizeCommand || command instanceof StatCommand)
		{
			this.indexConfig = new IndexConfig(this, command, compiler, props);
			if(command.needsResources())
			{
				runtime.addShutdownHook(new CloseDictionaries(this));
			}

			if(command instanceof SummarizeCommand)
			{
				this.summarizersConfig = new SummarizersConfig(this, (SummarizeCommand)command, compiler, props);
			}
		}
	}

	public void verifySystem() throws Exception
	{
		String[] checkStrings ={ DICTIONARIES_TRANSACTIONAL, LOAD_DELIM };

		for(int i=0; i<checkStrings.length; i++)
		{
			String s = props.getProperty(checkStrings[i]);
			if(s == null)
			{
				throw new Exception("The system setting '"+checkStrings[i]+"' is undefined");
			}
			else
			{
				//System.out.println("--"+checkStrings[i]+" = "+ s);
			}
		}

		String[] checkInts = {DICTIONARIES_COMMIT_EVERY, LOAD_READ_BUFFER_SIZE, LOAD_EXCEPTIONS_MAX,
		                      MAX_LOAD_FIELDS, INDEX_BLOCK_SIZE, MAX_EXPORT_FILE_SIZE,
		                      LOG_LEVEL, LOG_MODE};

		for(int i=0; i<checkInts.length; i++)
		{
			String s = props.getProperty(checkInts[i]);
			if(s == null)
			{
				throw new Exception("The system setting '"+checkInts[i]+"' is undefined");
			}
			else
			{
				int c = Integer.parseInt(s);
			}
		}
	}

}
