/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.system;

import amberstone.read.*;
import amberstone.file.*;
import amberstone.summary.*;

import java.io.*;
import java.util.*;

public class StatCommand implements Command
{
	private boolean valid;
	private String index;

	public StatCommand(String args[])
	{
		if(args.length == 2)
		{
			this.index = args[1];
			valid = true;
		}
		else
		{
			valid = false;
		}
	}

	public boolean valid()
	{
		return valid;
	}

	public String getIndex()
	{
		return this.index;
	}

	public String usage()
	{
		return "-stat [index]";
	}

	public void run(Config config)
	{
		try
		{
			RangeFilter fileFilter = new RangeFilter();
			Range range = new Range(10000101, 30001231);
			fileFilter.addRange(range);
			String dataWarehouse = config.getIndexConfig().getIndexDir();

			//System.out.println("Compiling stats on the data warehouse...");

			long records = 0L;
			long bytes = 0L;
			int blockCount = 0;
			Map<String, Long> columns = new HashMap();

			List<YearDir> years = YearDir.getYears(new File(dataWarehouse), fileFilter, new DirAsc());
			for(int y=0; y<years.size(); y++)
			{
				//System.out.println("Year:"+years.get(y).toString());
				List<MonthDir> months = years.get(y).getMonths();
				for(int m=0; m<months.size(); m++)
				{
					//System.out.println("Month:"+months.get(m).toString());

					List<DayDir> days = months.get(m).getDays();

					for(int d=0; d<days.size(); d++)
					{
						List<Block> blocks = days.get(d).getBlocks();

						for(int i=0; i<blocks.size(); i++)
						{
							Block block = blocks.get(i);
							File dir = block.getDir();
							File stats = new File(dir, "stats");
							BufferedReader in = new BufferedReader(new FileReader(stats));
							records += Integer.parseInt(in.readLine());
							in.close();
							++blockCount;
							File[] cols = dir.listFiles();
							for(int c=0; c<cols.length; c++)
							{
								File col = cols[c];
								if(col.getName().endsWith(".dat"))
								{
									String colName = col.getName().split("\\.")[0];
									bytes+= col.length();
									if(columns.containsKey(colName))
									{
										Long colBytes = columns.get(colName);
										long cb = colBytes.longValue();
										cb += col.length();
										columns.put(colName, new Long(cb));
									}
									else
									{
										Long colBytes = new Long(col.length());
										columns.put(colName, colBytes);
									}
								}
							}
						}
					}
				}
			}

			System.out.println("Total records:"+records);
			System.out.println("Total blocks:"+blockCount);
			System.out.println("Total MB:"+bytes/1000000);

			Iterator it = columns.keySet().iterator();
			System.out.println("Column MB:");
			while(it.hasNext())
			{
				String colName = (String)it.next();
				Long b = columns.get(colName);
				System.out.println("    "+colName+":"+b.longValue()/1000000);
			}
		}
		catch(Exception e)
		{
			System.out.println("Error in stat command:");
			System.out.println(e.getMessage());
			Trace trace = new Trace();
			trace.trace(e, "stat", config.getLogsHome());
		}
	}

	public boolean needsResources()
	{
		return true;
	}
}