/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.time;

import amberstone.lang.*;

import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Calendar;
import java.util.Arrays;
import java.util.Iterator;

public class WeekSeries extends TimeSeries
{

	private boolean incremental;

	private static final FunctionType[][] allow = {{FunctionType.STRING_LITERAL, FunctionType.STRING_LITERAL},
												   {FunctionType.WEEK, FunctionType.WEEK},
												   {FunctionType.STRING_LITERAL, FunctionType.WEEK},
												   {FunctionType.WEEK, FunctionType.STRING_LITERAL}};


	public FunctionType[][] strict()
	{
		return allow;
	}

	public WeekSeries(Week week)
	{
		increment = week;
		start = end = week.getIncrement();
	}

	public WeekSeries()
	{

	}

	public String output(String YYYYWW)
	{
		String out = YYYYWW.substring(4,6)+"-"+YYYYWW.substring(0, 4);
		return out;
	}

	public int granularity()
	{
		return 2;
	}

	public TimeSeries getTimeSeries(int start, int end)
	{
		WeekSeries weekSeries = new WeekSeries();
		weekSeries.start = start;
		weekSeries.end = end;
		return weekSeries;
	}

	public void params(Function[] params) throws ParameterException
	{
		try
		{
			start = validateAndFormat(params[0].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[0], "Invalid week format (YYYY-WW): "+params[0].toString(), e);
		}

		try
		{
			end = validateAndFormat(params[1].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[1], "Invalid week format (YYYY-WW): "+params[1].toString(), e);
		}

		if(start > end)
		{
			throw new ParameterException(params[0], "Start week must be less then or equal to end week");
		}

		if(params[1] instanceof Week)
		{
			this.incremental = true;
		}
	}

	public static int validateAndFormat(String time)
		throws Exception
	{
		String[] parts = time.split("-");

		if(parts.length != 2)
		{
			throw new Exception("Week must include two sections");
		}

		int YYYY = validateAndFormatYear(parts[0]);
		int WW = validateAndFormatWeek(parts[1]);
		return (YYYY*100)+WW;
	}

	public String toString()
	{
		return "weeks";
	}

	public int convertTo(TimeSeries convertSeries, int convertValue)
	{
		return convertSeries.convertToWeeks(convertValue);
	}

	public int convertToYears(int YYYYWW)
	{
		return (int)YYYYWW/100;
	}

	public int convertToMonths(int YYYYWW)
	{
		int YYYY = YYYYWW/100;
		int WW = YYYYWW%YYYY;

		cal.set(YYYY, 0, 1);
		while(true)
		{
			int testWW = cal.get(cal.WEEK_OF_YEAR);
			if(WW == testWW)
			{
				int MM = cal.get(cal.MONTH);
				return (YYYY*100)+MM;
			}
			else
			{
				cal.add(cal.DATE, 1);
			}
		}
	}

	public int convertToQuarters(int YYYYWW)
	{
		int YYYY = YYYYWW/100;
		int WW = YYYYWW%YYYY;

		cal.set(YYYY, 0, 1);
		while(true)
		{
			int testWW = cal.get(cal.WEEK_OF_YEAR);
			if(WW == testWW)
			{
				int MM = cal.get(cal.MONTH);
				MM++;
				switch(MM)
				{
					case 1:
						return (YYYY*10)+1;
					case 2:
						return (YYYY*10)+1;
					case 3:
						return (YYYY*10)+1;
					case 4:
						return (YYYY*10)+2;
					case 5:
						return (YYYY*10)+2;
					case 6:
						return (YYYY*10)+2;
					case 7:
						return (YYYY*10)+3;
					case 8:
						return (YYYY*10)+3;
					case 9:
						return (YYYY*10)+3;
					case 10:
						return (YYYY*10)+4;
					case 11:
						return (YYYY*10)+4;
					case 12:
						return (YYYY*10)+4;
				}
			}
			else
			{
				cal.add(cal.DATE, 1);
			}
		}
	}

	public int convertYYYYMMDD(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);
		int DD = YYYYMMDD % ((YYYY*10000)+(MM*100));
		cal.set(YYYY, MM-1, DD);
		int WW = cal.get(cal.WEEK_OF_YEAR);
		return (YYYY*100)+WW;
	}

	public int currentTime()
	{
		cal.setTimeInMillis(System.currentTimeMillis());
		int year  = cal.get(cal.YEAR);
		int week = cal.get(cal.WEEK_OF_YEAR);
		return (year*100)+week;
	}

	public int getTime(int YYYYWW)
	{
		int YYYY = (int)(YYYYWW/100);
		int WW = (int)(YYYYWW % (YYYY*100));
		cal.clear();
		cal.set(YYYY, 0, 1, 15, 0, 0);
		int week = cal.get(cal.WEEK_OF_YEAR);

		while(week != WW)
		{
			cal.add(cal.DATE, 7);
			week = cal.get(cal.WEEK_OF_YEAR);
		}

		return (int)cal.getTimeInMillis()/1000;
	}

	public int getMinYYYYMMDD()
	{
		int year = (int)(start/100);
		int week = start % (year*100);

		cal.set(year, 0,1);
		int testWeek = -1;

		while(true)
		{
			testWeek = cal.get(Calendar.WEEK_OF_YEAR);
			if(testWeek == week)
			{
				int YYYY = cal.get(Calendar.YEAR);
				int MM = cal.get(Calendar.MONTH);
				int DD = cal.get(Calendar.DAY_OF_MONTH);

				return (YYYY*10000)+((MM+1)*100)+DD;
			}
			else
			{
				cal.add(Calendar.DATE, 1);
			}
		}
	}

	public int[] getCols(int minYYYYMMDD, int maxYYYYMMDD)
	{

		int minYYYY = (int)(minYYYYMMDD/10000);
		int minMM = (int)((minYYYYMMDD % (minYYYY*10000)) / 100);
		int minDD = minYYYYMMDD % ((minYYYY*10000)+(minMM*100));

		Set<Integer> weeks = new HashSet();

		cal.set(minYYYY, minMM-1, minDD);
		int lastYear = -1;
		int lastWeek = -1;

		while(true)
		{
			int testWeek = cal.get(Calendar.WEEK_OF_YEAR);
			int testDay = cal.get(Calendar.DAY_OF_WEEK);


			int testYYYY = cal.get(Calendar.YEAR);
			int testMM = cal.get(Calendar.MONTH);
			int testDD = cal.get(Calendar.DAY_OF_MONTH);

			int testYYYYMMDD = (testYYYY*10000)+((testMM+1)*100)+testDD;

			if(testYYYYMMDD <= maxYYYYMMDD)
			{
				//System.out.println("Hello:"+testYYYY +" "+ testWeek);
				if(testWeek < lastWeek && testYYYY <= lastYear)
				{
					testYYYY++;
				}

				weeks.add(new Integer((testYYYY*100)+testWeek));
				cal.add(Calendar.DATE, 1);
				lastYear = testYYYY;
				lastWeek = testWeek;
			}
			else
			{
				break;
			}
		}
		int[] y = new int[weeks.size()];

		Iterator<Integer> it = weeks.iterator();
		int i = -1;
		while(it.hasNext())
		{
			y[++i] = it.next().intValue();
		}

		Arrays.sort(y);

		return y;
	}

	public int getMaxYYYYMMDD()
	{
		int year = (int)(end/100);
		int week = end % (year*100);

		cal.set(year, 0,1);
		int testWeek = -1;

		while(true)
		{
			testWeek = cal.get(Calendar.WEEK_OF_YEAR);
			int testDay = cal.get(Calendar.DAY_OF_WEEK);
			int testYear = cal.get(Calendar.YEAR);


			if(testWeek == week && testDay == 7)
			{
				int YYYY = cal.get(Calendar.YEAR);
				int MM = cal.get(Calendar.MONTH);
				int DD = cal.get(Calendar.DAY_OF_MONTH);

				return (YYYY*10000)+((MM+1)*100)+DD;
			}
			else if(testWeek == week && testYear > year)
			{
				cal.add(Calendar.DATE, -1);
				int YYYY = cal.get(Calendar.YEAR);
				int MM = cal.get(Calendar.MONTH);
				int DD = cal.get(Calendar.DAY_OF_MONTH);
				return (YYYY*10000)+((MM+1)*100)+DD;
			}
			else
			{
				cal.add(Calendar.DATE, 1);
			}
		}
	}
}