/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.transaction;

import amberstone.data.Key;
import amberstone.util.Primes;

import java.util.List;
import java.util.ArrayList;

public class TransactionLRU
{
	private TransactionQueue lru = new TransactionQueue();
	private Transaction transactions[];
	private int size;
	private int resize;
	int count;

	public TransactionLRU(int size)
	{
		int s = (int)(size*1.15);
		this.size = Primes.selectPrime(s);
		this.resize = (int)(size*.75);
		transactions = new Transaction[this.size];
	}

	public Transaction peekLRU()
	{
		return lru.peekLast();
	}

	public void remove(Transaction transaction)
	{
		Key key   = transaction.id;
		int index = key.hashCode() % this.size;
		Transaction t = null;
		Transaction last = null;
		for(t = transactions[index]; t != null; last = t, t = t.link)
		{
			if(key.equalsKey(t.id))
			{
				if(last == null)
				{
					transactions[index] = t.link;
				}
				else
				{
					last.link = t.link;
				}
				lru.remove(t);
				--count;
				return;
			}
		}
	}

	public void put(Transaction transaction)
	{
		Key key   = transaction.id;
		int index = key.hashCode() % this.size;
		Transaction t = null;
		Transaction last= null;
		for(t = transactions[index]; t != null; last = t, t = t.link);

		if(last == null)
		{
			transactions[index] = transaction;
			lru.addFirst(transaction);
			count++;
		}
		else
		{
			last.link = transaction;
			lru.addFirst(transaction);
			count++;
		}

		if(count > resize)
		{
			resize();
		}

		return;
	}

	public void rehash(Transaction transaction)
	{
		transaction.link = null;
		Key key   = transaction.id;
		int index = key.hashCode() % this.size;
		Transaction t = null;
		Transaction last= null;
		for(t = transactions[index]; t != null; last = t, t = t.link);

		if(last == null)
		{
			transactions[index] = transaction;
		}
		else
		{
			last.link = transaction;
		}
	}

	private List<Transaction> getTransactions()
	{
		List<Transaction> list = new ArrayList(count);
		for(int i=0; i<transactions.length; i++)
		{
			Transaction t = transactions[i];
			if(t != null)
			{
				list.add(t);
				for(t = t.link; t != null; t = t.link)
				{
					list.add(t);
				}
			}
		}

		return list;
	}

	private void resize()
	{
		List<Transaction> list = getTransactions();
		int s = (int)(size*1.75);
		size = Primes.selectPrimeCandidate(s);
		resize = (int)(size*.75);
		transactions = new Transaction[size];
		count = 0;
		for(int i=0; i<list.size(); i++)
		{
			Transaction t = list.get(i);
			rehash(t);
		}
	}

	public Transaction get(Key key)
	{
		int index = key.hashCode() % this.size;
		for(Transaction t = transactions[index]; t != null; t = t.link)
		{
			if(key.equalsKey(t.id))
			{
				lru.moveFirst(t);
				return t;
			}
		}

		return null;
	}
}