/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.transaction;

import amberstone.data.Key;
import amberstone.lang.*;
import amberstone.file.FileStat;
import amberstone.summary.*;

import java.util.List;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.Calendar;
import java.io.File;
import java.io.IOException;

public class TransactionTracker implements FileNode
{
	private TransactionTrackerTemplate transactionTrackerTemplate;

	private KeyFunction idFunc;
	private TimeFunction timeFunc;
	private int size;
	private int timeout;

	private TransactionLRU lru;
	private FileNode[] fileNodes;
	private PropTemplate[] props;
	private Calendar cal = new GregorianCalendar();

	public TransactionTracker(TransactionTrackerTemplate transactionTrackerTemplate,
						      KeyFunction idFunc,
						      TimeFunction timeFunc,
						      int timeout,
						      int size)
	{
		this.transactionTrackerTemplate = transactionTrackerTemplate;
		this.idFunc                     = idFunc;
		this.timeout                    = timeout;
		this.timeFunc                   = timeFunc;
		this.size                       = size;
		this.lru                        = new TransactionLRU(size);
	}

	public TransactionTracker(TransactionTrackerTemplate transactionTrackerTemplate,
						      KeyFunction idFunc,
						      TimeFunction timeFunc,
						      int timeout,
						      TransactionLRU activeLRU)
	{
		this.transactionTrackerTemplate = transactionTrackerTemplate;
		this.idFunc                     = idFunc;
		this.timeout                    = timeout;
		this.timeFunc                   = timeFunc;
		this.size                       = size;
		this.lru                        = activeLRU;
	}

	public void setFileNodes(FileNode[] fileNodes)
	{
		this.fileNodes = fileNodes;
	}

	public FileNode[] getFileNodes()
	{
		return this.fileNodes;
	}

	public void setProps(PropTemplate[] props)
	{
		this.props = props;
	}

	public void output() throws Exception
	{
		for(int i=0; i<fileNodes.length; i++)
		{
			fileNodes[i].output();
		}
	}

	public String getName()
	{
		return null;
	}

	public void update() throws FunctionException
	{
		Key id = idFunc.getKey();
		int currentTime = timeFunc.getTime();

		Transaction t = this.lru.get(id);

		if(t == null)
		{
			t = new Transaction(id.cloneKey(), currentTime);
			t.setProps(getProps(props));
			t.update(currentTime);
			flushTransactions(currentTime);
			lru.put(t);
		}
		else
		{
			t.update(currentTime);
		}
	}

	public void finish(long grain) throws FunctionException
	{
		flushTransactions(getTime(grain));
		for(int i=0; i<fileNodes.length; i++)
		{
			fileNodes[i].finish(grain);
		}

		this.transactionTrackerTemplate.activeLRU = lru;
	}

	private int getTime(long grain)
	{
		int dayGrain = ActiveYearMonthDay.ACTIVE_KEY.intValue();
		if(grain > 2147483647)
		{
			int hourGrain = (int)grain;
			int YYYY = (int)(dayGrain/10000);
			int MM = (int)((dayGrain % (YYYY*10000)) / 100);
			int DD = dayGrain%((YYYY*10000)+(MM*100));
			cal.clear();
			cal.set(YYYY, MM-1, DD, 0, 0, 0);
			if(this.transactionTrackerTemplate.currentDay == dayGrain || this.transactionTrackerTemplate.currentDay == 0)
			{
				cal.add(cal.HOUR, hourGrain+1);
			}

			this.transactionTrackerTemplate.currentDay = dayGrain;
		}
		else
		{
			int YYYY = (int)(dayGrain/10000);
			int MM = (int)((dayGrain % (YYYY*10000)) / 100);
			int DD = dayGrain%((YYYY*10000)+(MM*100));
			cal.clear();
			cal.set(YYYY, MM-1, DD, 0, 0, 0);
		}

		long t = cal.getTimeInMillis();
		return (int)(t/1000);
	}

	private void flushTransactions(int currentTime) throws FunctionException
	{
		while(true)
		{
			Transaction oldTransaction = lru.peekLRU();

			if(oldTransaction != null && (currentTime - oldTransaction.lastTouched) > timeout)
			{
				lru.remove(oldTransaction);
				this.transactionTrackerTemplate.transaction = oldTransaction;

				for(int i=0; i<fileNodes.length; i++)
				{
					fileNodes[i].update();
				}
			}
			else
			{
				break;
			}
		}
	}

	private Prop[] getProps(PropTemplate[] ps) throws FunctionException
	{
		Prop[] newProps = new Prop[ps.length];
		for(int i=0; i<ps.length; i++)
		{
			newProps[i] = ps[i].getProp();
		}

		return newProps;
	}
}